
import threading
class OutputPlugin(object):
  def __init__(self, name):
    self.__name = name
    self._size = 0
    self.__limit_size = 1000
    self._queue = []
    self._queue_mutex = threading.Lock()
    self._enable_query = False
  def Init(self, options):
    pass
  def GetName(self):
    return self.__name
  def Put(self, target, timestamp, keys, values):
    record = self.GenerateRecord(target, timestamp, keys, values)
    with self._queue_mutex:
      if self._size >= self.__limit_size:
        self.LoseRecord()
      self._queue.append(record)
      self._size += 1
  def Pop(self):
    pass
  def LoseRecord(self):
    pass
  def GenerateRecord(self, target, timestamp, keys, values):
    pass
  def Query(self, json_request):
    pass
  def EnableQuery(self):
    return self._enable_query




class MetricException(Exception):
  def __init__(self, info):
    super(MetricException, self).__init__()
    self.except_info = info
  def __str__(self):
    return "MetricException:" + self.except_info



from .dbotmetrics_collector import DBotMetricsCollector
from .base_metric import BaseMetric
from .metric_exception import MetricException
from .version import __version__

__all__ = ('DBotMetricsCollector', 'BaseMetric', 'MetricException')




import configparser

class Configuration():
  def __init__(self, config_file):
    self.__collector_file = config_file
    self.__config_parser = configparser.ConfigParser()
    self.__config_parser.read(self.__collector_file)

    self.__metric_buffer_limit = 0
    self.__precision = 0
    self.__plugins = {}
    self.__parse_sections()
  def Get(self, section, key):
    return self.__config_parser.get(section, key)
  def GetMetricBufferLimit(self):
    return self.__metric_buffer_limit
  def GetPluginOptions(self):
    return self.__plugins
  def __get_options(self, section):
    options = self.__config_parser.options(section)
    opts = {}
    for opt in options:
      opts[opt] = self.__config_parser.get(section, opt)
    return opts
      
  def __parse_sections(self):
    self.__metric_buffer_limit = self.__config_parser.get('base', 'metric_buffer_limit')
    self.__precision = self.__config_parser.get('base', 'precision')

    sections = self.__config_parser.sections()
    for sec in sections:
      secsplit = sec.split('.', 1)
      if len(secsplit) != 2:
        continue
      sectype = secsplit[0]
      secname = secsplit[1]
      if self.__plugins.get(sectype) is None:
        self.__plugins[sectype] = {secname: self.__get_options(sec)}
      else:
        self.__plugins[sectype][secname] = self.__get_options(sec)
        



class BaseMetric(object):
  def __init__(self):
    pass
  def Init(self, options):
    raise NotImplemented()
  '''
  @return String
  '''
  def GetName(self):
    raise NotImplemented()
  '''
  @para notify: a callback handle, with one parameter: metric records
  '''
  def RegisterNotify(self, notify):
    raise NotImplemented()
  '''
  @para flag: Donot record every time the metric collected if False
  '''
  def EnableDetailRecord(self, flag):
    raise NotImplemented()
  '''
  @return a list of metric records, with format like flowing
          [{'target': 'table_student', 'keys': {"name":"steven", "no": 11223344}, 'values': {"age": 23}}]
  '''
  def FetchInfo(self):
    raise NotImplemented()
  '''
  @return the collecting interval(seconds)
  '''
  def GetCollectInterval(self):
    raise NotImplemented()




import datetime
import logging
import threading
import json
import os
from .timer import Timer
from .configuation import Configuration
from .metric_exception import MetricException

class SingleInstance(object):
  _instance = None
  def __new__(cls, *args, **kw):
    if not cls._instance:
      cls._instance = super(SingleInstance, cls).__new__(cls, *args, **kw)  
    return cls._instance 

class DBotMetricsCollector(SingleInstance):
  _is_inited = False
  def __init__(self):
    if self._is_inited:
      return
    self._is_inited = True
    self.__mutex_metrics = threading.Lock()
    self.__metrics = {}
    self.__mutex_output = threading.Lock()
    self.__output_plugins = {}
    self.__unit_time = 1
    self.__timer = Timer(self.__unit_time, self.__on_timer)
    self.__config_file = None
    self.__config = None
  def Start(self, config=None):
    if config:
      self.__config_file = config
    else:
      self.__config_file = '/etc/collector.conf'

    if not os.path.isfile(self.__config_file):
      raise MetricException("Please set config file collector.conf or copy it to default path /etc")

    self.__config = Configuration(config)
    logging.info("Metric dispatcher start, config %s", self.__config_file)
    self.__load_plugins()
    self.__timer.start()
  def Stop(self):
    logging.info("Metric dispatcher stop")
    self.__timer.stoptimer()

  def __load_plugins(self):
    plugins = self.__config.GetPluginOptions()
    for plugin_type, plugin_options in plugins.items():
      for name, options in plugin_options.items():
        enable = options.get('enable')
        if enable == 'True':
          try:
            pluginfile = options.get('file')
            plugin_module = options.get('module')
            plugin = __import__("dbot_metrics." + plugin_type + "." + pluginfile, fromlist=[plugin_module])
            objclass = getattr(plugin, plugin_module)
            o = objclass()
            o.Init(options)
            self.__register_plugin(plugin_type, o)
          except:
            logging.exception("import plugin failed")

  def __on_timer(self):
    with self.__mutex_metrics:
      for _, m in self.__metrics.items():
        m.time_counter += 1
        if m.time_counter >= m.GetCollectInterval():
          infos = m.FetchInfo()
          if len(infos) == 0:
            m.time_counter = 0
            continue
          lastinfo = infos.pop()
          for info in infos:
            self.__process_metric(info, False)
          self.__process_metric(lastinfo, True)
          m.time_counter = 0

  def __process_metric(self, data, push_immediate):
    timestamp = float(datetime.datetime.utcnow().strftime("%s.%f"))
    with self.__mutex_output:
      for _, output in self.__output_plugins.items():
        output.Put(data['target'], timestamp, data['keys'], data['values'])
        if push_immediate:
          output.Push()

  def __register_plugin(self, plugin_type, plugin_obj):
    if plugin_type == "output":
      self.__register_output_plugin(plugin_obj)
    elif plugin_type == "input":
      raise MetricException("Not support input plugins")
      #self.RegisterMetric(plugin_obj)

  def RegisterMetric(self, metric):
    logging.info("Metric register metric plugin %s", metric.GetName())
    with self.__mutex_metrics:
      self.__metrics[metric.GetName()] = metric
    metric.RegisterNotify(self.__process_metric)

  def GetMetric(self, name):
    with self.__mutex_metrics:
      return self.__metrics.get(name)

  def __register_output_plugin(self, output):
    logging.info("Metric register output plugin %s", output.GetName())
    with self.__mutex_output:
      self.__output_plugins[output.GetName()] = output

  def __unregister_output_plugin(self, name):
    logging.info("Metric unregister output plugin %s", name)
    with self.__mutex_output:
      del self.__output_plugins[name]

  def Query(self, request):
    result = []
    with self.__mutex_output:
      for name, output in self.__output_plugins.items():
        if output.EnableQuery():
          result.extend([{'name':name, 'content':output.Query(request)}])
    return result




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dbot_metrics",
    version="0.1.5",
    author="xiaoch05",
    author_email="xiaoch2010@gmail.com",
    description="ATN atn dbot metrics package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ATNIO/metrics",
    packages = setuptools.find_packages(),
    install_requires=["leveldb >= 0.194"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)




from collections import MutableMapping
import leveldb
import json
from .output_plugin import OutputPlugin
import threading
import logging
import datetime
import os
from datetime import timedelta

class LevelDBDict(MutableMapping):
  def __init__(self, db_file):
    self.__db_file = db_file
    self.__db = leveldb.LevelDB(self.__db_file)
  def __len__(self):
    return len(self.keys())
  def __getitem__(self, key):
    k = str.encode(str(key)) if not isinstance(key, bytes) else key
    return self.__db.Get(k)

  def __setitem__(self, key, value):
    k = str.encode(str(key)) if not isinstance(key, bytes) else key
    self.__db.Put(k, value)

  def __delitem__(self, key):
    k = str.encode(str(key)) if not isinstance(key, bytes) else key
    self.__db.Delete(k)

  def __iter__(self):
    for k in self.__db.RangeIter(include_value=False):
      yield k

  def keys(self):
    return [k for k, v in self.__db.RangeIter()]

  def items(self):
    return self.__db.RangeIter()

  def rangescan(self, start=None, end=None):
    if start is None and end is None:
      return self.__db.RangeIter()
    elif end is None:
      k = str.encode(str(start)) if not isinstance(start, bytes) else start
      return self.__db.RangeIter(k)
    else:
      start = str.encode(str(start)) if not isinstance(start, bytes) else start
      end = str.encode(str(end)) if not isinstance(end, bytes) else end
      return self.__db.RangeIter(start, end)

  def writebatch(self, values):
    batch = leveldb.WriteBatch()
    for key, value in values.items():
      k = str.encode(str(key)) if not isinstance(key, bytes) else key
      batch.Put(k, str.encode(json.dumps(value)))
    self.__db.Write(batch)

  def deleterange(self, start, end):
    batch = leveldb.WriteBatch()
    iterator = self.rangescan(start, end)
    for k in iterator:
      batch.Delete(k[0])
    self.__db.Write(batch)

class LevelDBJsonDict(LevelDBDict):
    def __getitem__(self, key):
        return json.loads(LevelDBDict.__getitem__(self, key))

    def __setitem__(self, key, value):
        LevelDBDict.__setitem__(self, key, str.encode(json.dumps(value)))

    def iteritems(self):
        for k, v in LevelDBDict.iteritems(self):
            yield k, json.loads(v)

    def rangescan(self, start=None, end=None):
        for k, v in LevelDBDict.rangescan(self, start, end):
            yield k, json.loads(v)

class LevelDBClient(OutputPlugin):
  def __init__(self):
    super(LevelDBClient, self).__init__("leveldb")
    self.__cache = []
    self.__cache_mutex = threading.Lock()
    self.__history = None
    self.__datapath = None
    self.__database = {}
  def Init(self, options):
    database = options.get("database")
    enable_query = options.get('enable_query')
    self._enable_query = (enable_query == 'True')
    history = options.get("history")
    self.__history = {'s': lambda x: timedelta(seconds=x),
                      'm': lambda x: timedelta(minutes=x),
                      'h': lambda x: timedelta(hours=x),
                      'd': lambda x: timedelta(days=x),
                      'y': lambda x: timedelta(years=x)}[history[-1]](int(history[:-1]))
    logging.info("configure-- database %s", database)
    if not os.path.exists(database):
      os.makedirs(database)
    self.__datapath = database

  def Push(self):
    try:
      with self.__cache_mutex:
        while len(self.__cache) > 0:
          cache = self.__cache[0]
          cache['table'].writebatch(cache['records'])
          self._size -= len(cache['records'])
          self.__cache.pop(0)
    except:
        logging.exception("leveldb send cache failed")
        return
    
    records = {}
    with self._queue_mutex:
      for body in self._queue:
        record = { 'time':body["time"], 'tags':body['tags'], 'fields':body['fields'] }

        table_name = body['table']
        recordlist = records.get(table_name)
        if recordlist:
          recordlist['records'][record['time']] = record
        else:
          table = self.__database.get(table_name, None)
          if table is None:
            table = LevelDBJsonDict(self.__datapath + "/" + table_name + ".db")
            self.__database[table_name] = table
          records[table_name] = {'table':table, 'records':{record['time']: record}}
        
      self._queue[:] = []
    try:
      tables={}
      for r in list(records.keys()):
        record = records[r]
        record['table'].writebatch(record['records'])
        self._size -= len(record['records'])
        tables[r] = record['table']
        del records[r]
      rmv_before = int((datetime.datetime.utcnow() - self.__history).strftime("%s"))
      for _, t in tables.items():
        self.RmvOlder(t, rmv_before)
    except:
      with self.__cache_mutex:
        for _, record in records.items():
          self.__cache.append(record)
  def GenerateRecord(self, table, time, tags, fields):
    json_body = {"table":table, "tags": tags, "time": time, "fields": fields}
    return json_body
  def LoseRecord(self):
    with self.__cache_mutex:
      if len(self.__cache) > 0:
        first = self.__cache[0]
        del(first[0])
        self._size -= 1
        if len(first) == 0:
          del(self.__cache[0])
      elif len(self._queue) > 0:
        del(self._queue[0])
        self._size -= 1

  def Query(self, json_request):
    ts_range = json_request.get('range')
    ts_from = ts_range.get('from')
    ts_to = ts_range.get('to')
    targets = json_request.get('targets')
    for target in targets:
      table_name = target.get('table')
      table = self.__database.get(table_name, None) 
      if table is None:
        table = LevelDBJsonDict(self.__datapath + "/" + table_name + ".db")
        self.__database[table_name] = table
      query_type = target.get('type')
      if query_type == 'all':
        return [v[1] for v in list(table.rangescan(ts_from, ts_to))]
      else:
        logging.error("Unsupported query type %s", query_type)
        return None

  def RmvOlder(self, t, oldtime):
    t.deleterange(0, oldtime)
    



from threading import Thread
from threading import Event

class Timer(Thread):
    def __init__(self, interval, function, args=[], kwargs={}):
        Thread.__init__(self)
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = Event()
 
    def stoptimer(self):
        self.finished.set()
 
    def run(self):
        while True:
            self.finished.wait(self.interval)
            if self.finished.is_set():
                self.finished.set()
                break
            self.function(*self.args, **self.kwargs) 
